@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: CDDL-1.0
 /*
  * CDDL HEADER START
  *
@@ -608,29 +609,23 @@ get_replication(nvlist_t *nvroot, boolean_t fatal)
 				verify(nvlist_lookup_string(cnv,
 				    ZPOOL_CONFIG_PATH, &path) == 0);
 
-				/*
-				 * Skip active spares they should never cause
-				 * the pool to be evaluated as inconsistent.
-				 */
-				if (is_spare(NULL, path))
-					continue;
-
 				/*
 				 * If we have a raidz/mirror that combines disks
-				 * with files, only report it as an error when
-				 * fatal is set to ensure all the replication
-				 * checks aren't skipped in check_replication().
+				 * with files, report it as an error.
 				 */
-				if (fatal && !dontreport && type != NULL &&
+				if (!dontreport && type != NULL &&
 				    strcmp(type, childtype) != 0) {
 					if (ret != NULL)
 						free(ret);
 					ret = NULL;
-					vdev_error(gettext(
-					    "mismatched replication "
-					    "level: %s contains both "
-					    "files and devices\n"),
-					    rep.zprl_type);
+					if (fatal)
+						vdev_error(gettext(
+						    "mismatched replication "
+						    "level: %s contains both "
+						    "files and devices\n"),
+						    rep.zprl_type);
+					else
+						return (NULL);
 					dontreport = B_TRUE;
 				}
 
