@@ -1,3 +1,4 @@
+// SPDX-License-Identifier: CDDL-1.0
 /*
  * CDDL HEADER START
  *
@@ -32,7 +33,6 @@
 #include <linux/exportfs.h>
 #include <linux/falloc.h>
 #include <linux/parser.h>
-#include <linux/task_io_accounting_ops.h>
 #include <linux/vfs_compat.h>
 #include <linux/writeback.h>
 #include <linux/xattr_compat.h>
@@ -123,41 +123,6 @@ extern int zpl_clone_file_range(struct file *src_file, loff_t src_off,
 extern int zpl_dedupe_file_range(struct file *src_file, loff_t src_off,
     struct file *dst_file, loff_t dst_off, uint64_t len);
 
-/* compat for FICLONE/FICLONERANGE/FIDEDUPERANGE ioctls */
-typedef struct {
-	int64_t		fcr_src_fd;
-	uint64_t	fcr_src_offset;
-	uint64_t	fcr_src_length;
-	uint64_t	fcr_dest_offset;
-} zfs_ioc_compat_file_clone_range_t;
-
-typedef struct {
-	int64_t		fdri_dest_fd;
-	uint64_t	fdri_dest_offset;
-	uint64_t	fdri_bytes_deduped;
-	int32_t		fdri_status;
-	uint32_t	fdri_reserved;
-} zfs_ioc_compat_dedupe_range_info_t;
-
-typedef struct {
-	uint64_t	fdr_src_offset;
-	uint64_t	fdr_src_length;
-	uint16_t	fdr_dest_count;
-	uint16_t	fdr_reserved1;
-	uint32_t	fdr_reserved2;
-	zfs_ioc_compat_dedupe_range_info_t	fdr_info[];
-} zfs_ioc_compat_dedupe_range_t;
-
-#define	ZFS_IOC_COMPAT_FICLONE		_IOW(0x94, 9, int)
-#define	ZFS_IOC_COMPAT_FICLONERANGE \
-    _IOW(0x94, 13, zfs_ioc_compat_file_clone_range_t)
-#define	ZFS_IOC_COMPAT_FIDEDUPERANGE \
-    _IOWR(0x94, 54, zfs_ioc_compat_dedupe_range_t)
-
-extern long zpl_ioctl_ficlone(struct file *filp, void *arg);
-extern long zpl_ioctl_ficlonerange(struct file *filp, void *arg);
-extern long zpl_ioctl_fideduperange(struct file *filp, void *arg);
-
 
 #if defined(HAVE_INODE_TIMESTAMP_TRUNCATE)
 #define	zpl_inode_timestamp_truncate(ts, ip)	timestamp_truncate(ts, ip)
